import pandas as pd


class GraphDataFrame(pd.DataFrame):
    """DataFrame subclass for CPI indexes."""

    def node(self, node):
        """
        Retrieve a node's series from the DataFrame.

        Parameters
        ----------
        node : str
            The node to retrieve.

        Returns
        -------
        pd.Series
            The series corresponding to the node.
        """
        return self[node]

    def children(self, node, max_depth=False, strict=True):
        """
        Retrieve the children of a node.

        Parameters
        ----------
        node : str
            The parent node.
        max_depth : bool, optional
            If True, return all levels of children. If False, return only the next level (default is True).
        strict : bool, optional
            If True, the parent node itself will not be included (default is False).

        Returns
        -------
        pd.DataFrame
            DataFrame containing the children of the node.
        """
        if node == "CP00":
            # If the parent node is CP00, children should be CP01 to CP12
            if max_depth:
                if strict:
                    mask = self.columns.isin(
                        [code for code in self.columns if code.startswith("CP") and code != "CP00"]
                    )
                else:
                    return self
            else:
                if strict:
                    mask = self.columns.isin(["CP" + str(i).zfill(2) for i in range(1, 13)])
                else:
                    mask = self.columns.isin(["CP" + str(i).zfill(2) for i in range(0, 13)])
        else:
            if max_depth:
                mask = self.columns.str.startswith(node)
            else:
                mask = self.columns.str.startswith(node) & (self.columns.str.len() <= len(node) + 1)
            if strict:
                mask = mask & (self.columns != node)

        return self.loc[:, mask]


class CountryGraphDataFrame(pd.DataFrame):
    """DataFrame subclass for multi-country CPI indexes."""

    def country(self, country):
        """Retrieve data for a specific country."""
        return self[country]

    def node(self, node, country=None):
        """
        Retrieve a node's series for a specific country.

        Parameters
        ----------
        node : str
            The node to retrieve.
        country : str, optional
            The country for which to retrieve the node. If None, retrieves the node for all countries.

        Returns
        -------
        pd.Series or pd.DataFrame
            The series corresponding to the node for a specific country,
            or a DataFrame containing the series for the node for all countries.
        """
        if country is None:
            return self.xs(node, level=1, axis=1)
        return self[country, node]

    def children(self, node, country, max_depth=False, strict=True):
        """
        Retrieve the children of a node for a specific country.

        Parameters
        ----------
        node : str
            The parent node.
        country : str
            The country for which to retrieve the children.
        max_depth : bool, optional
            If True, return all levels of children. If False, return only the next level (default is True).
        strict : bool, optional
            If True, the parent node itself will not be included (default is False).

        Returns
        -------
        pd.DataFrame
            DataFrame containing the children of the node for the specified country.
        """
        return GraphDataFrame(self[country]).children(node, max_depth, strict)


class CPI:
    """CPI system."""

    def __init__(self, prices, weights, country, base_year):
        """Initialize the CPI object for a single country."""
        self.prices = GraphDataFrame(prices)
        self.weights = GraphDataFrame(weights)
        self.country = country
        self.base_year = base_year


class CPICountries:
    """Multi-county CPI system."""

    def __init__(self, prices, weights, country_weights, base_year):
        """Initialize the CPI object for multiple countries."""
        self.prices = CountryGraphDataFrame(prices)
        self.weights = CountryGraphDataFrame(weights)
        self.country_weights = country_weights
        self.base_year = base_year



import gzip
import os
import time
import urllib.request

import pandas as pd

# pylint: disable = broad-exception-caught, invalid-name, unspecified-encoding


def file_age(filename):
    """
    Calculate the age of a given file in hours.

    Parameters
    ----------
    filename : str
        The name of the file.

    Returns
    -------
    int
        The age of the file in hours (rounded to whole hours), -1 if file does not exist.
    """
    if os.path.exists(filename):
        return int((time.time() - os.path.getmtime(filename)) / 3600)
    return -1


def download_url(url, filename, binary=False, unzip=False):
    """
    Download the given URL and save it under filename.
    If the filename contains a directory, it is assumed the directory exists.

    Parameters
    ----------
    url : str
        The URL to download.
    filename : str
        The filename to save the file under.
    binary : bool, optional
        If True, download file as binary.
    unzip : bool, optional
        If True, unzip the downloaded file.
    """
    if os.path.exists(filename) and file_age(filename) < 24:
        print("Cached version exists and is less than a day old for", filename)
        return
    try:
        with urllib.request.urlopen(url) as resp:
            if unzip:
                with gzip.GzipFile(fileobj=resp) as data:
                    file_content = data.read()
            else:
                file_content = resp.read()
            if not binary:
                file_content = file_content.decode("utf-8")

        dir_name = os.path.dirname(filename)
        os.makedirs(dir_name, exist_ok=True)

        with open(filename, "wb" if binary else "w") as f:
            f.write(file_content)
    except Exception as e:
        print(e)


def get_eurostat_dictionary(dictionary, inverse=False):
    """
    Fetch a dictionary from Eurostat.

    Parameters
    ----------
    dictionary : str
        The name of the dictionary to download.
    inverse : bool, optional
        If True, return value -> key mapping, defaults to False.

    Returns
    -------
    dict
        A Python dictionary with the key -> value pair.
    """
    dictionary = dictionary.lower()
    url = (
        "https://ec.europa.eu/eurostat/estat-navtree-portlet-prod/"
        "BulkDownloadListing?sort=1&downfile=dic%2Fen%2F{dictionary}.dic"
    )
    filename = os.path.join("cache", f"{dictionary}.dic")
    download_url(url, filename)

    try:
        with open(filename) as f:
            d = {}
            for line in f:
                if len(line.strip()) > 1:
                    row = line.split("\t")
                    d[row[0]] = row[1].strip()
        return {v: k for k, v in d.items()} if inverse else d
    except Exception as e:
        print(e)
        return {}


def get_eurostat_dataset(dataset, replace_codes=True, transpose=True, keep_codes=None):
    """
    Fetch a dataset from Eurostat.

    Downloads the dataset, replaces the code columns with their associated meanings (optional) and
    transposes the column for improved usability (optional).

    Parameters
    ----------
    dataset : str
        The name of the dataset to download.
    replace_codes : bool, optional
        If True, replaces codes with their associated values. Defaults to True.
    transpose : bool, optional
        If True, transpose the    dataset for ease of analysis. Defaults to True.
    keep_codes : list, optional
        List of codes not to replace. Defaults to None.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the downloaded dataset.
    """
    dataset = dataset.lower()
    filename = os.path.join("cache", f"{dataset}.tsv")
    url = f"https://ec.europa.eu/eurostat/estat-navtree-portlet-prod/BulkDownloadListing?file=data/{dataset}.tsv.gz"
    download_url(url, filename, unzip=True)

    df = pd.read_csv(filename, sep=",|\t| [^ ]?\t", na_values=":", engine="python", dtype_backend="pyarrow")
    df.columns = [x.split("\\")[0].strip() for x in df.columns]

    # Get dictionary columns
    with open(os.path.join("cache", f"{dataset}.tsv")) as f:
        first_line = f.readline()
    codes = first_line.split("\t")[0].split("\\")[0].split(",")

    # Replace codes with value
    if replace_codes:
        for c in codes:
            if keep_codes is None or c not in keep_codes:
                code_list = get_eurostat_dictionary(c)
                df[c] = df[c].replace(code_list)

    # Transpose the table
    if transpose:
        df = df.set_index(codes).transpose()

    return df


def get_eurostat_geodata(level=0):
    """
    Fetch the geodata of the European countries.

    The features are filtered on the NUTS level code. Level 0 contains the countries,
    level 1 the regions within countries, etc.

    Parameters
    ----------
    level : int, optional
        The NUTS level to download. Defaults to 0.

    Returns
    -------
    GeoDataFrame
        A GeoDataFrame with the geodata.
    """
    import geopandas as gpd  # pylint: disable = import-outside-toplevel

    url = "https://gisco-services.ec.europa.eu/distribution/v2/nuts/shp/NUTS_RG_20M_2021_3035.shp.zip"
    filename = os.path.join("cache", "NUTS_RG_20M_2021_3035.shp.zip")
    download_url(url, filename, binary=True)

    borders = gpd.read_file(filename)
    return borders[borders["LEVL_CODE"] == level]



import pandas as pd

from cpilib.cpi.base_classes import CPICountries
from cpilib.utils import get_eurostat_dataset


class HICP(CPICountries):
    """CPI class for the Harmonized Index of Consumer Prices (HICP)."""

    def __init__(self):
        prices = get_eurostat_dataset("prc_hicp_midx")
        prices.index = pd.to_datetime(prices.index, format="%YM%m")
        prices = prices["I15"].swaplevel(0, 1, axis=1).sort_index(axis=1).sort_index(axis=0)

        item_weights = get_eurostat_dataset("prc_hicp_inw")
        item_weights.index = pd.to_datetime(item_weights.index)
        item_weights = item_weights.swaplevel(0, 1, axis=1).sort_index(axis=1).sort_index(axis=0)

        country_weights = get_eurostat_dataset("prc_hicp_cow")
        country_weights.index = pd.to_datetime(country_weights.index)
        country_weights = country_weights["COWEA19"].sort_index(axis=0)

        super().__init__(prices, item_weights, country_weights, 2015)
